from collections.abc import Sequence
from typing import Any, Optional

import boto3.session
import dagster._check as check

from dagster_aws.utils import construct_boto_client_retry_config


def construct_ssm_client(
    max_attempts: int,
    region_name: Optional[str] = None,
    profile_name: Optional[str] = None,
    endpoint_url: Optional[str] = None,
    use_ssl: bool = True,
    aws_access_key_id: Optional[str] = None,
    aws_secret_access_key: Optional[str] = None,
    aws_session_token: Optional[str] = None,
    verify: Optional[bool] = None,
):
    check.int_param(max_attempts, "max_attempts")
    check.opt_str_param(region_name, "region_name")
    check.opt_str_param(profile_name, "profile_name")
    check.opt_str_param(endpoint_url, "endpoint_url")
    check.bool_param(use_ssl, "use_ssl")
    check.opt_bool_param(verify, "verify")
    check.opt_str_param(aws_access_key_id, "aws_access_key_id")
    check.opt_str_param(aws_secret_access_key, "aws_secret_access_key")
    check.opt_str_param(aws_session_token, "aws_session_token")

    client_session = boto3.session.Session(profile_name=profile_name)
    ssm_client = client_session.client(
        "ssm",
        region_name=region_name,
        use_ssl=use_ssl,
        verify=verify,
        endpoint_url=endpoint_url,
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
        aws_session_token=aws_session_token,
        config=construct_boto_client_retry_config(max_attempts),
    )

    return ssm_client


def get_parameters_by_tags(
    ssm_manager, parameter_tags: Sequence[dict[str, Any]], with_decryption: bool
) -> dict[str, str]:
    """Return a dictionary of AWS Secrets Manager names to arns
    for any secret tagged with `secrets_tag`.
    """
    parameter_names = []
    paginator = ssm_manager.get_paginator("describe_parameters")
    for parameter_tag in parameter_tags:
        filter_spec = {
            "Key": f"tag:{parameter_tag['key']}",
        }
        if parameter_tag.get("values"):
            filter_spec.update(Values=parameter_tag["values"])

        for page in paginator.paginate(
            ParameterFilters=[filter_spec],
        ):
            for param in page["Parameters"]:
                parameter_names.append(param["Name"])
    if not parameter_names:
        return {}
    else:
        return get_parameters_by_name(ssm_manager, parameter_names, with_decryption)


def get_parameters_by_name(
    ssm_manager, parameter_names: list[str], with_decryption: bool
) -> dict[str, str]:
    """Return a dictionary of AWS Parameter Store parameter names and their values."""
    parameter_values = {}
    for retrieved in ssm_manager.get_parameters(
        Names=parameter_names, WithDecryption=with_decryption
    )["Parameters"]:
        parameter_values[retrieved["Name"]] = retrieved["Value"]

    return parameter_values


def get_parameters_by_paths(
    ssm_manager, parameter_paths: list[dict[str, str]], with_decryption: bool, recursive: bool
) -> dict[str, str]:
    """Returns a dictionary of AWS Parameter Store parameter names and their values that match a list of paths. If
    recursive == True, then return all parameters that are prefixed by the given path.
    """
    parameter_values = {}
    for path in parameter_paths:
        paginator = ssm_manager.get_paginator("get_parameters_by_path")
        for page in paginator.paginate(
            Path=path, Recursive=recursive, WithDecryption=with_decryption
        ):
            for parameter in page["Parameters"]:
                parameter_values[parameter["Name"]] = parameter["Value"]
    return parameter_values



import json
import os
import random
import string
from collections.abc import Iterator
from contextlib import contextmanager
from typing import TYPE_CHECKING

import boto3
import dagster._check as check
from dagster._core.pipes.client import PipesContextInjector, PipesParams
from dagster._core.pipes.utils import PipesEnvContextInjector

if TYPE_CHECKING:
    from dagster_pipes import PipesContextData

_CONTEXT_FILENAME = "context.json"


class PipesS3ContextInjector(PipesContextInjector):
    """A context injector that injects context by writing to a temporary S3 location.

    Args:
        bucket (str): The S3 bucket to write to.
        client (boto3.client): A boto3 client to use to write to S3.
        key_prefix (Optional[str]): An optional prefix to use for the S3 key. Defaults to a random
            string.

    """

    def __init__(self, *, bucket: str, client: boto3.client):  # pyright: ignore (reportGeneralTypeIssues)
        super().__init__()
        self.bucket = check.str_param(bucket, "bucket")
        self.client = client

    @contextmanager
    def inject_context(self, context: "PipesContextData") -> Iterator[PipesParams]:
        key_prefix = "".join(random.choices(string.ascii_letters, k=30))
        key = os.path.join(key_prefix, _CONTEXT_FILENAME)
        self.client.put_object(
            Body=json.dumps(context).encode("utf-8"), Bucket=self.bucket, Key=key
        )
        yield {"bucket": self.bucket, "key": key}
        self.client.delete_object(Bucket=self.bucket, Key=key)

    def no_messages_debug_text(self) -> str:
        return (
            "Attempted to inject context via a temporary file in s3. Expected"
            " PipesS3ContextLoader to be explicitly passed to open_dagster_pipes in the external"
            " process."
        )


class PipesLambdaEventContextInjector(PipesEnvContextInjector):
    """Injects context via AWS Lambda event input.
    Should be paired with :py:class`~dagster_pipes.PipesMappingParamsLoader` on the Lambda side.
    """

    def no_messages_debug_text(self) -> str:
        return "Attempted to inject context via the lambda event input."



from collections.abc import Generator
from contextlib import contextmanager
from typing import TYPE_CHECKING, Any, Optional, cast

from dagster import (
    Config,
    Field as LegacyDagsterField,
    resource,
)
from dagster._annotations import beta
from dagster._config.field_utils import Shape
from dagster._core.definitions.resource_definition import dagster_maintained_resource
from dagster._core.test_utils import environ
from dagster._utils.merger import merge_dicts
from pydantic import Field

from dagster_aws.ssm.parameters import (
    construct_ssm_client,
    get_parameters_by_name,
    get_parameters_by_paths,
    get_parameters_by_tags,
)
from dagster_aws.utils import ResourceWithBoto3Configuration

if TYPE_CHECKING:
    import botocore


@beta
class SSMResource(ResourceWithBoto3Configuration):
    """Resource that gives access to AWS Systems Manager Parameter Store.

    The underlying Parameter Store session is created by calling
    :py:func:`boto3.session.Session(profile_name) <boto3:boto3.session>`.
    The returned resource object is a Systems Manager client, an instance of `botocore.client.ssm`.

    Example:
        .. code-block:: python

            from typing import Any
            from dagster import build_op_context, job, op
            from dagster_aws.ssm import SSMResource

            @op
            def example_ssm_op(ssm: SSMResource):
                return ssm.get_client().get_parameter(
                    Name="a_parameter"
                )

            @job
            def example_job():
                example_ssm_op()

            defs = Definitions(
                jobs=[example_job],
                resources={
                    'ssm': SSMResource(
                        region_name='us-west-1'
                    )
                }
            )
    """

    @classmethod
    def _is_dagster_maintained(cls) -> bool:
        return True

    def get_client(self) -> "botocore.client.ssm":  # pyright: ignore (reportAttributeAccessIssue)
        return construct_ssm_client(
            max_attempts=self.max_attempts,
            region_name=self.region_name,
            profile_name=self.profile_name,
            endpoint_url=self.endpoint_url,
            use_ssl=self.use_ssl,
            verify=self.verify,
            aws_access_key_id=self.aws_access_key_id,
            aws_secret_access_key=self.aws_secret_access_key,
            aws_session_token=self.aws_session_token,
        )


@beta
@dagster_maintained_resource
@resource(config_schema=SSMResource.to_config_schema())
def ssm_resource(context) -> "botocore.client.ssm":  # pyright: ignore (reportAttributeAccessIssue)
    """Resource that gives access to AWS Systems Manager Parameter Store.

    The underlying Parameter Store session is created by calling
    :py:func:`boto3.session.Session(profile_name) <boto3:boto3.session>`.
    The returned resource object is a Systems Manager client, an instance of `botocore.client.ssm`.

    Example:
        .. code-block:: python

            from dagster import build_op_context, job, op
            from dagster_aws.ssm import ssm_resource

            @op(required_resource_keys={'ssm'})
            def example_ssm_op(context):
                return context.resources.ssm.get_parameter(
                    Name="a_parameter"
                )

            @job(resource_defs={'ssm': ssm_resource})
            def example_job():
                example_ssm_op()

            example_job.execute_in_process(
                run_config={
                    'resources': {
                        'ssm': {
                            'config': {
                                'region_name': 'us-west-1',
                            }
                        }
                    }
                }
            )

    You may configure this resource as follows:

    .. code-block:: YAML

        resources:
          parameter_store:
            config:
              region_name: "us-west-1"
              # Optional[str]: Specifies a custom region for the Parameter Store session. Default is chosen
              # through the ordinary boto credential chain.
              profile_name: "dev"
              # Optional[str]: Specifies a custom profile for Parameter Store session. Default is default
              # profile as specified in ~/.aws/credentials file

    """
    return SSMResource.from_resource_context(context).get_client()


@beta
class ParameterStoreTag(Config):
    key: str = Field(description="Tag key to search for.")
    values: Optional[list[str]] = Field(default=None, description="List")


@beta
class ParameterStoreResource(ResourceWithBoto3Configuration):
    """Resource that provides a dict which maps selected SSM Parameter Store parameters to
    their string values. Optionally sets selected parameters as environment variables.

    Example:
        .. code-block:: python

            import os
            from typing import Dict
            from dagster import build_op_context, job, op
            from dagster_aws.ssm import ParameterStoreResource, ParameterStoreTag

            @op
            def example_parameter_store_op(parameter_store: ParameterStoreResource):
                return parameter_store.fetch_parameters().get("my-parameter-name")

            @op
            def example_parameter_store_op_2(parameter_store: ParameterStoreResource):
                with parameter_store.parameters_in_environment():
                    return os.getenv("my-other-parameter-name")

            @job
            def example_job():
                example_parameter_store_op()
                example_parameter_store_op_2()

            defs = Definitions(
                jobs=[example_job],
                resource_defs={
                    'parameter_store': ParameterStoreResource(
                        region_name='us-west-1',
                        parameter_tags=[ParameterStoreTag(key='my-tag-key', values=['my-tag-value'])],
                        add_to_environment=True,
                        with_decryption=True,
                    )
                },
            )
    """

    parameters: list[str] = Field(
        default=[], description="An array of AWS SSM Parameter Store parameter names to fetch."
    )
    parameter_tags: list[ParameterStoreTag] = Field(
        default=[],
        description=(
            "AWS SSM Parameter store parameters with this tag will be fetched and made available."
        ),
    )
    parameter_paths: list[str] = Field(
        default=[], description="List of path prefixes to pull parameters from."
    )
    with_decryption: bool = Field(
        default=False,
        description=(
            "Whether to decrypt parameters upon retrieval. Is ignored by AWS if parameter type is"
            " String or StringList"
        ),
    )

    @classmethod
    def _is_dagster_maintained(cls) -> bool:
        return True

    @contextmanager
    def parameters_in_environment(
        self,
        parameters: Optional[list[str]] = None,
        parameter_tags: Optional[list[ParameterStoreTag]] = None,
        parameter_paths: Optional[list[str]] = None,
    ) -> Generator[dict[str, str], None, None]:
        """Yields a dict which maps selected Parameter Store parameters to their string values. Also
        sets chosen parameters as environment variables.

        Args:
            parameters (Optional[List[str]]):   An array of AWS SSM Parameter Store parameter names to fetch.
                Note that this will override the parameters specified in the resource config.
            parameter_tags (Optional[List[ParameterStoreTag]]): AWS SSM Parameter store parameters with this tag
                will be fetched and made available. Note that this will override the parameter_tags specified
                in the resource config.
            parameter_paths (Optional[List[str]]): List of path prefixes to pull parameters from. Note that this
                will override the parameter_paths specified in the resource config.
        """
        ssm_manager = construct_ssm_client(
            max_attempts=self.max_attempts,
            region_name=self.region_name,
            profile_name=self.profile_name,
        )
        parameters_to_fetch = parameters if parameters is not None else self.parameters
        parameter_tags_to_fetch = (
            parameter_tags if parameter_tags is not None else self.parameter_tags
        )
        parameter_paths_to_fetch = (
            parameter_paths if parameter_paths is not None else self.parameter_paths
        )

        results = []
        if parameters_to_fetch:
            results.append(
                get_parameters_by_name(ssm_manager, parameters_to_fetch, self.with_decryption)
            )
        if parameter_tags_to_fetch:
            parameter_tag_inputs = [
                {"key": tag.key, "values": tag.values} for tag in parameter_tags_to_fetch
            ]
            results.append(
                get_parameters_by_tags(ssm_manager, parameter_tag_inputs, self.with_decryption)
            )
        if parameter_paths_to_fetch:
            results.append(
                get_parameters_by_paths(
                    ssm_manager,
                    parameter_paths_to_fetch,  # type: ignore
                    self.with_decryption,
                    recursive=True,
                )
            )
        if not results:
            parameter_values = {}
        else:
            if len(results) > 1:
                parameter_values = merge_dicts(*results)
            else:
                parameter_values = results[0]

        with environ(parameter_values):
            yield parameter_values

    def fetch_parameters(
        self,
        parameters: Optional[list[str]] = None,
        parameter_tags: Optional[list[ParameterStoreTag]] = None,
        parameter_paths: Optional[list[str]] = None,
    ) -> dict[str, str]:
        """Fetches parameters from SSM Parameter Store and returns them as a dict.

        Args:
            parameters (Optional[List[str]]):   An array of AWS SSM Parameter Store parameter names to fetch.
                Note that this will override the parameters specified in the resource config.
            parameter_tags (Optional[List[ParameterStoreTag]]): AWS SSM Parameter store parameters with this tag
                will be fetched and made available. Note that this will override the parameter_tags specified
                in the resource config.
            parameter_paths (Optional[List[str]]): List of path prefixes to pull parameters from. Note that this
                will override the parameter_paths specified in the resource config.
        """
        with self.parameters_in_environment(
            parameters=parameters, parameter_tags=parameter_tags, parameter_paths=parameter_paths
        ) as parameter_values:
            return parameter_values


LEGACY_PARAMETERSTORE_SCHEMA = {
    **cast(Shape, ParameterStoreResource.to_config_schema().as_field().config_type).fields,
    "add_to_environment": LegacyDagsterField(
        bool,
        default_value=False,
        description="Whether to add the parameters to the environment. Defaults to False.",
    ),
}


@beta
@dagster_maintained_resource
@resource(config_schema=LEGACY_PARAMETERSTORE_SCHEMA)
@contextmanager
def parameter_store_resource(context) -> Any:
    """Resource that provides a dict which maps selected SSM Parameter Store parameters to
    their string values. Optionally sets selected parameters as environment variables.

    Example:
        .. code-block:: python

            import os
            from dagster import build_op_context, job, op
            from dagster_aws.ssm import parameter_store_resource

            @op(required_resource_keys={'parameter_store'})
            def example_parameter_store_op(context):
                return context.resources.parameter_store.get("my-parameter-name")

            @op(required_resource_keys={'parameter_store'})
            def example_parameter_store_op_2(context):
                return os.getenv("my-other-parameter-name")

            @job(resource_defs={'parameter_store': parameter_store_resource})
            def example_job():
                example_parameter_store_op()
                example_parameter_store_op_2()

            example_job.execute_in_process(
                run_config={
                    'resources': {
                        'parameter_store': {
                            'config': {
                                'region_name': 'us-west-1',
                                'parameter_tags': 'dagster',
                                'add_to_environment': True,
                                'with_decryption': True,
                            }
                        }
                    }
                }
            )

    Note that your ops must also declare that they require this resource with
    `required_resource_keys`, or it will not be initialized for the execution of their compute
    functions.

    You may configure this resource as follows:

    .. code-block:: YAML

        resources:
          parameter_store:
            config:
              region_name: "us-west-1"
              # Optional[str]: Specifies a custom region for the Parameter Store session. Default is chosen
              # through the ordinary boto credential chain.
              profile_name: "dev"
              # Optional[str]: Specifies a custom profile for Parameter Store session. Default is default
              # profile as specified in ~/.aws/credentials file
              parameters: ["parameter1", "/path/based/parameter2"]
              # Optional[List[str]]: Specifies a list of parameter names to pull from parameter store.
              parameters_tag: "dagster"
              # Optional[Sequence[Dict[str, Any]]]: Specifies a list of tag specifications, all parameters which have the tag set
              will be pulled  from Parameter Store. Each tag specification is in the format {"tag": "tag name or prefix", "option": "BeginsWith|Equals"};
              when option == "BeginsWith", all parameters with tags that start with the tag value will be pulled.
              add_to_environment: true
              # Optional[bool]: Whether to set the selected parameters as environment variables. Defaults
              # to false.

    """
    add_to_environment = context.resource_config.get("add_to_environment", False)
    if add_to_environment:
        with ParameterStoreResource.from_resource_context(
            context
        ).parameters_in_environment() as secrets:
            yield secrets
    else:
        yield ParameterStoreResource.from_resource_context(context).fetch_parameters()



import sys
import time
from typing import TYPE_CHECKING, Any, Optional, Union, cast

import boto3
import dagster._check as check
from dagster import DagsterInvariantViolationError, PipesClient
from dagster._annotations import public
from dagster._core.definitions.metadata import RawMetadataMapping
from dagster._core.definitions.resource_annotation import TreatAsResourceParam
from dagster._core.errors import DagsterExecutionInterruptedError
from dagster._core.execution.context.asset_execution_context import AssetExecutionContext
from dagster._core.execution.context.compute import OpExecutionContext
from dagster._core.pipes.client import (
    PipesClientCompletedInvocation,
    PipesContextInjector,
    PipesMessageReader,
)
from dagster._core.pipes.context import PipesSession
from dagster._core.pipes.utils import PipesEnvContextInjector, open_pipes_session
from dagster._utils.merger import deep_merge_dicts

from dagster_aws.pipes.message_readers import PipesCloudWatchLogReader, PipesCloudWatchMessageReader

if TYPE_CHECKING:
    from mypy_boto3_emr_serverless.client import EMRServerlessClient
    from mypy_boto3_emr_serverless.literals import JobRunStateType
    from mypy_boto3_emr_serverless.type_defs import (
        GetJobRunResponseTypeDef,
        MonitoringConfigurationTypeDef,
        StartJobRunRequestRequestTypeDef,
        StartJobRunResponseTypeDef,
    )

AWS_SERVICE_NAME = "EMR Serverless"


@public
class PipesEMRServerlessClient(PipesClient, TreatAsResourceParam):
    """A pipes client for running workloads on AWS EMR Serverless.

    Args:
        client (Optional[boto3.client]): The boto3 AWS EMR Serverless client used to interact with AWS EMR Serverless.
        context_injector (Optional[PipesContextInjector]): A context injector to use to inject
            context into AWS EMR Serverless workload. Defaults to :py:class:`PipesEnvContextInjector`.
        message_reader (Optional[PipesMessageReader]): A message reader to use to read messages
            from the AWS EMR Serverless workload. Defaults to :py:class:`PipesCloudWatchMessageReader`.
        forward_termination (bool): Whether to cancel the AWS EMR Serverless workload if the Dagster process receives a termination signal.
        poll_interval (float): The interval in seconds to poll the AWS EMR Serverless workload for status updates. Defaults to 5 seconds.
    """

    AWS_SERVICE_NAME = AWS_SERVICE_NAME

    def __init__(
        self,
        client: Optional["EMRServerlessClient"] = None,
        context_injector: Optional[PipesContextInjector] = None,
        message_reader: Optional[PipesMessageReader] = None,
        forward_termination: bool = True,
        poll_interval: float = 5.0,
    ):
        self._client: EMRServerlessClient = cast(
            "EMRServerlessClient", client or boto3.client("emr-serverless")
        )
        self._context_injector = context_injector or PipesEnvContextInjector()
        self._message_reader = message_reader or PipesCloudWatchMessageReader()
        self.forward_termination = check.bool_param(forward_termination, "forward_termination")
        self.poll_interval = poll_interval

    @property
    def client(self) -> "EMRServerlessClient":
        return self._client

    @property
    def context_injector(self) -> PipesContextInjector:
        return self._context_injector

    @property
    def message_reader(self) -> PipesMessageReader:
        return self._message_reader

    @classmethod
    def _is_dagster_maintained(cls) -> bool:
        return True

    @public
    def run(
        self,
        *,
        context: Union[OpExecutionContext, AssetExecutionContext],
        start_job_run_params: "StartJobRunRequestRequestTypeDef",
        extras: Optional[dict[str, Any]] = None,
    ) -> PipesClientCompletedInvocation:
        """Run a workload on AWS EMR Serverless, enriched with the pipes protocol.

        Args:
            context (Union[OpExecutionContext, AssetExecutionContext]): The context of the currently executing Dagster op or asset.
            params (dict): Parameters for the ``start_job_run`` boto3 AWS EMR Serverless client call.
                See `Boto3 EMR Serverless API Documentation <https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/emr-serverless/client/start_job_run.html>`_
            extras (Optional[Dict[str, Any]]): Additional information to pass to the Pipes session in the external process.

        Returns:
            PipesClientCompletedInvocation: Wrapper containing results reported by the external
            process.
        """
        with open_pipes_session(
            context=context,
            message_reader=self.message_reader,
            context_injector=self.context_injector,
            extras=extras,
        ) as session:
            start_job_run_params = self._enrich_start_params(context, session, start_job_run_params)
            start_response = self._start(context, start_job_run_params)
            try:
                completion_response = self._wait_for_completion(context, start_response)
                context.log.info(f"[pipes] {self.AWS_SERVICE_NAME} workload is complete!")
                self._read_messages(context, session, completion_response)
                return PipesClientCompletedInvocation(
                    session, metadata=self._extract_dagster_metadata(completion_response)
                )

            except DagsterExecutionInterruptedError:
                if self.forward_termination:
                    context.log.warning(
                        f"[pipes] Dagster process interrupted! Will terminate external {self.AWS_SERVICE_NAME} workload."
                    )
                    self._terminate(context, start_response)
                raise

    def _enrich_start_params(
        self,
        context: Union[OpExecutionContext, AssetExecutionContext],
        session: PipesSession,
        params: "StartJobRunRequestRequestTypeDef",
    ) -> "StartJobRunRequestRequestTypeDef":
        # inject Dagster tags
        tags = params.get("tags", {})
        params["tags"] = {**tags, **session.default_remote_invocation_info}

        # inject env variables via --conf spark.executorEnv.env.<key>=<value>
        dagster_env_vars = {}

        dagster_env_vars.update(session.get_bootstrap_env_vars())

        if "jobDriver" not in params:
            params["jobDriver"] = {}

        if "sparkSubmit" not in params["jobDriver"]:
            params["jobDriver"]["sparkSubmit"] = {}  # pyright: ignore[reportGeneralTypeIssues]

        params["jobDriver"]["sparkSubmit"]["sparkSubmitParameters"] = params.get(
            "jobDriver", {}
        ).get("sparkSubmit", {}).get("sparkSubmitParameters", "") + "".join(
            [
                f" --conf spark.emr-serverless.driverEnv.{key}={value}"
                for key, value in dagster_env_vars.items()
            ]
        )

        return cast("StartJobRunRequestRequestTypeDef", params)

    def _start(
        self,
        context: Union[OpExecutionContext, AssetExecutionContext],
        params: "StartJobRunRequestRequestTypeDef",
    ) -> "StartJobRunResponseTypeDef":
        response = self.client.start_job_run(**params)
        job_run_id = response["jobRunId"]

        context.log.info(
            f"[pipes] {self.AWS_SERVICE_NAME} job started with job_run_id {job_run_id}."
        )

        return response

    def _wait_for_completion(
        self,
        context: Union[OpExecutionContext, AssetExecutionContext],
        start_response: "StartJobRunResponseTypeDef",
    ) -> "GetJobRunResponseTypeDef":  # pyright: ignore[reportReturnType]
        job_run_id = start_response["jobRunId"]
        application_id = start_response["applicationId"]

        running_dashboard_url = None
        completed_dashboard_url = None

        while response := self.client.get_job_run(
            applicationId=start_response["applicationId"],
            jobRunId=job_run_id,
        ):
            state: JobRunStateType = response["jobRun"]["state"]

            # get dashboard url when it's ready (but only once)
            if state == "RUNNING" and running_dashboard_url is None:
                running_dashboard_url = self.client.get_dashboard_for_job_run(
                    applicationId=application_id, jobRunId=job_run_id
                )
                context.log.info(
                    f"[pipes] {self.AWS_SERVICE_NAME} job is running. Dashboard URL: {running_dashboard_url}"
                )
            # completed jobs have a different dashboard url
            elif state in ["SUCCEEDED", "FAILED", "CANCELLED"] and completed_dashboard_url is None:
                completed_dashboard_url = self.client.get_dashboard_for_job_run(
                    applicationId=application_id, jobRunId=job_run_id
                )
                context.log.info(
                    f"[pipes] {self.AWS_SERVICE_NAME} job is completed. Dashboard URL: {completed_dashboard_url}"
                )

            # check if the job is in a terminal state
            if state in ["FAILED", "CANCELLED", "CANCELLING"]:
                context.log.error(
                    f"[pipes] {self.AWS_SERVICE_NAME} job {job_run_id} terminated with state: {state}. Details:\n{response['jobRun'].get('stateDetails')}"
                )
                raise RuntimeError(
                    f"{self.AWS_SERVICE_NAME} job failed"
                )  # TODO: introduce something like DagsterPipesRemoteExecutionError
            elif state == "SUCCESS":
                context.log.info(
                    f"[pipes] {self.AWS_SERVICE_NAME} job {job_run_id} completed with state: {state}"
                )
                return response
            elif state in ["PENDING", "SUBMITTED", "SCHEDULED", "RUNNING", "QUEUED"]:
                time.sleep(self.poll_interval)
                continue
            else:
                raise DagsterInvariantViolationError(
                    f"Unexpected state for AWS EMR Serverless job {job_run_id}: {state}"
                )

    def _read_messages(
        self,
        context: Union[OpExecutionContext, AssetExecutionContext],
        session: PipesSession,
        response: "GetJobRunResponseTypeDef",
    ):
        application_id = response["jobRun"]["applicationId"]
        job_id = response["jobRun"]["jobRunId"]

        application = self.client.get_application(applicationId=application_id)["application"]

        # merge base monitoring configuration from application
        # with potential overrides from the job run
        application_monitoring_configuration = application.get("monitoringConfiguration", {})
        job_monitoring_configuration = (
            response["jobRun"].get("configurationOverrides", {}).get("monitoringConfiguration", {})
        )
        monitoring_configuration = cast(
            "MonitoringConfigurationTypeDef",
            deep_merge_dicts(application_monitoring_configuration, job_monitoring_configuration),
        )

        application_type = application["type"]

        if application_type == "Spark":
            worker_type = "SPARK_DRIVER"
        elif application_type == "Hive":
            worker_type = "HIVE_DRIVER"
        else:
            raise NotImplementedError(f"Application type {application_type} is not supported")

        if not isinstance(self.message_reader, PipesCloudWatchMessageReader):
            context.log.warning(
                f"[pipes] {self.message_reader} is not supported for {self.AWS_SERVICE_NAME}. Dagster won't be able to receive logs and messages from the job."
            )
            return

        # https://docs.aws.amazon.com/emr/latest/EMR-Serverless-UserGuide/logging.html#jobs-log-storage-cw

        # we can get cloudwatch logs from the known log group

        if (
            monitoring_configuration.get("cloudWatchLoggingConfiguration", {}).get("enabled")
            is not True
        ):
            context.log.warning(
                f"[pipes] Recieved {self.message_reader}, but CloudWatch logging is not enabled for {self.AWS_SERVICE_NAME} job. Dagster won't be able to receive logs and messages from the job."
            )
            return

        if log_types := monitoring_configuration.get("cloudWatchLoggingConfiguration", {}).get(
            "logTypes"
        ):
            # get the configured output streams
            # but limit them with "stdout" and "stderr"
            output_streams = list(
                map(
                    lambda x: x.lower(),
                    set(log_types.get(worker_type, ["STDOUT", "STDERR"])) & {"stdout", "stderr"},
                )
            )
        else:
            output_streams = ["stdout", "stderr"]

        log_group = monitoring_configuration.get("logGroupName") or "/aws/emr-serverless"

        attempt = response["jobRun"].get("attempt")

        if attempt is not None and attempt > 1:
            log_stream = (
                f"/applications/{application_id}/jobs/{job_id}/attempts/{attempt}/{worker_type}"
            )
        else:
            log_stream = f"/applications/{application_id}/jobs/{job_id}/{worker_type}"

        if log_stream_prefix := monitoring_configuration.get(
            "cloudWatchLoggingConfiguration", {}
        ).get("logStreamNamePrefix"):
            log_stream = f"{log_stream_prefix}{log_stream}"

        output_files = {
            "stdout": sys.stdout,
            "stderr": sys.stderr,
        }

        # update MessageReader params so it can start receiving messages
        if isinstance(self.message_reader, PipesCloudWatchMessageReader):
            session.report_launched(
                {
                    "extras": {
                        "log_group": log_group,
                        "log_stream": f"{log_stream}/stdout",
                    }
                }
            )

            # now add LogReaders for stdout and stderr logs
            for output_stream in output_streams:
                output_file = output_files[output_stream]
                context.log.debug(
                    f"[pipes] Adding PipesCloudWatchLogReader for group {log_group} stream {log_stream}/{output_stream}"
                )
                self.message_reader.add_log_reader(
                    PipesCloudWatchLogReader(
                        client=self.message_reader.client,
                        log_group=log_group,
                        log_stream=f"{log_stream}/{output_stream}",
                        target_stream=output_file,
                        start_time=int(session.created_at.timestamp() * 1000),
                        debug_info=output_stream,
                    ),
                )

    def _extract_dagster_metadata(self, response: "GetJobRunResponseTypeDef") -> RawMetadataMapping:
        metadata: RawMetadataMapping = {}

        job_run = response["jobRun"]

        metadata["AWS EMR Serverless Application ID"] = job_run["applicationId"]
        metadata["AWS EMR Serverless Job Run ID"] = job_run["jobRunId"]

        # TODO: it would be great to add a url to EMR Studio page for this run
        # such urls look like: https://es-638xhdetxum2td9nc3a45evmn.emrstudio-prod.eu-north-1.amazonaws.com/#/serverless-applications/00fm4oe0607u5a1d
        # but we need to get the Studio ID from the application_id
        # which is not possible with the current AWS API

        return metadata

    def _terminate(
        self,
        context: Union[OpExecutionContext, AssetExecutionContext],
        start_response: "StartJobRunResponseTypeDef",
    ):
        job_run_id = start_response["jobRunId"]
        application_id = start_response["applicationId"]
        context.log.info(f"[pipes] Terminating {self.AWS_SERVICE_NAME} job run {job_run_id}")
        self.client.cancel_job_run(applicationId=application_id, jobRunId=job_run_id)



from collections.abc import Mapping, Sequence
from typing import TYPE_CHECKING, Any, NamedTuple, Optional, cast

from dagster import (
    Array,
    BoolSource,
    Field,
    Noneable,
    Permissive,
    Shape,
    StringSource,
    _check as check,
)
from dagster._config import process_config
from dagster._core.container_context import process_shared_container_context_config
from dagster._core.errors import DagsterInvalidConfigError
from dagster._core.storage.dagster_run import DagsterRun
from dagster._core.utils import parse_env_var

from dagster_aws.secretsmanager import get_tagged_secrets

if TYPE_CHECKING:
    from dagster_aws.ecs import EcsRunLauncher

# Config shared between EcsRunLauncher and EcsContainerContext
SHARED_ECS_SCHEMA = {
    "env_vars": Field(
        [StringSource],
        is_required=False,
        description=(
            "List of environment variable names to include in the ECS task. "
            "Each can be of the form KEY=VALUE or just KEY (in which case the value will be pulled "
            "from the current process)"
        ),
    ),
    "run_resources": Field(
        Permissive(
            {
                "cpu": Field(
                    str,
                    is_required=False,
                    description="The CPU override to use for the launched task.",
                ),
                "memory": Field(
                    str,
                    is_required=False,
                    description="The memory override to use for the launched task.",
                ),
                "ephemeral_storage": Field(
                    int,
                    is_required=False,
                    description="The ephemeral storage, in GiB, to use for the launched task.",
                ),
            }
        )
    ),
    "run_ecs_tags": Field(
        Array(
            {
                "key": Field(StringSource, is_required=True),
                "value": Field(StringSource, is_required=False),
            }
        ),
        is_required=False,
        description="Additional tags to apply to the launched ECS task.",
    ),
}

SHARED_TASK_DEFINITION_FIELDS = {
    "execution_role_arn": Field(
        StringSource,
        is_required=False,
        description=(
            "ARN of the task execution role for the ECS container and Fargate agent to make AWS API"
            " calls on your behalf. See"
            " https://docs.aws.amazon.com/AmazonECS/latest/developerguide/task_execution_IAM_role.html. "
        ),
    ),
    "task_role_arn": Field(
        StringSource,
        is_required=False,
        description=(
            "ARN of the IAM role for launched tasks. See"
            " https://docs.aws.amazon.com/AmazonECS/latest/developerguide/task-iam-roles.html. "
        ),
    ),
    "runtime_platform": Field(
        Shape(
            {
                "cpuArchitecture": Field(StringSource, is_required=False),
                "operatingSystemFamily": Field(StringSource, is_required=False),
            }
        ),
        is_required=False,
        description=(
            "The operating system that the task definition is running on. See"
            " https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ecs.html#ECS.Client.register_task_definition"
            " for the available options."
        ),
    ),
    "volumes": Field(
        Array(
            Permissive(
                {
                    "name": Field(StringSource, is_required=False),
                }
            )
        ),
        is_required=False,
        description=(
            "List of data volume definitions for the task. See"
            " https://docs.aws.amazon.com/AmazonECS/latest/developerguide/efs-volumes.html"
            " for the full list of available options."
        ),
    ),
    "mount_points": Field(
        Array(
            Shape(
                {
                    "sourceVolume": Field(StringSource, is_required=False),
                    "containerPath": Field(StringSource, is_required=False),
                    "readOnly": Field(BoolSource, is_required=False),
                }
            )
        ),
        is_required=False,
        description=(
            "Mount points for data volumes in the main container of the task."
            " See https://docs.aws.amazon.com/AmazonECS/latest/developerguide/efs-volumes.html"
            " for more information."
        ),
    ),
    "repository_credentials": Field(
        StringSource,
        is_required=False,
        description=(
            "The arn of the secret to authenticate into your private container registry."
            " This does not apply if you are leveraging ECR for your images, see"
            " https://docs.aws.amazon.com/AmazonECS/latest/userguide/private-auth.html."
        ),
    ),
}

ECS_CONTAINER_CONTEXT_SCHEMA = {
    "secrets": Field(
        Noneable(Array(Shape({"name": StringSource, "valueFrom": StringSource}))),
        is_required=False,
        description=(
            "An array of AWS Secrets Manager secrets. These secrets will "
            "be mounted as environment variables in the container. See "
            "https://docs.aws.amazon.com/AmazonECS/latest/APIReference/API_Secret.html."
        ),
    ),
    "secrets_tags": Field(
        Noneable(Array(StringSource)),
        is_required=False,
        description=(
            "AWS Secrets Manager secrets with these tags will be mounted as "
            "environment variables in the container."
        ),
    ),
    "task_definition_arn": Field(
        StringSource,
        is_required=False,
        description="ARN of the task definition to use to launch the container.",
    ),
    "container_name": Field(
        StringSource,
        is_required=False,
        description="Name of the container in the task definition to use to run Dagster code.",
    ),
    "server_resources": Field(
        Permissive(
            {
                "cpu": Field(
                    str,
                    is_required=False,
                    description="The CPU override to use for the launched task.",
                ),
                "memory": Field(
                    str,
                    is_required=False,
                    description="The memory override to use for the launched task.",
                ),
                "ephemeral_storage": Field(
                    int,
                    is_required=False,
                    description="The ephemeral storage, in GiB, to use for the launched task.",
                ),
                "replica_count": Field(
                    int,
                    default_value=1,
                    is_required=False,
                    description="The number of code server instances to launch.",
                ),
            }
        )
    ),
    "server_sidecar_containers": Field(
        Array(Permissive({})),
        is_required=False,
        description="Additional sidecar containers to include in code server task definitions.",
    ),
    "server_ecs_tags": Field(
        Array(
            {
                "key": Field(StringSource, is_required=True),
                "value": Field(StringSource, is_required=False),
            }
        ),
        is_required=False,
        description="Additional tags to apply to the launched ECS task for a code server.",
    ),
    "run_sidecar_containers": Field(
        Array(Permissive({})),
        is_required=False,
        description="Additional sidecar containers to include in run task definitions.",
    ),
    "server_health_check": Field(
        Permissive(),
        is_required=False,
        description="Health check to include in code server task definitions.",
    ),
    **SHARED_TASK_DEFINITION_FIELDS,
    **SHARED_ECS_SCHEMA,
}


class EcsContainerContext(
    NamedTuple(
        "_EcsContainerContext",
        [
            ("secrets", Sequence[Any]),
            ("secrets_tags", Sequence[str]),
            ("env_vars", Sequence[str]),
            ("task_definition_arn", Optional[str]),
            ("container_name", Optional[str]),
            ("server_resources", Mapping[str, Any]),
            ("run_resources", Mapping[str, Any]),
            ("task_role_arn", Optional[str]),
            ("execution_role_arn", Optional[str]),
            ("runtime_platform", Mapping[str, Any]),
            ("mount_points", Sequence[Mapping[str, Any]]),
            ("volumes", Sequence[Mapping[str, Any]]),
            ("server_sidecar_containers", Sequence[Mapping[str, Any]]),
            ("run_sidecar_containers", Sequence[Mapping[str, Any]]),
            ("server_ecs_tags", Sequence[Mapping[str, Optional[str]]]),
            ("run_ecs_tags", Sequence[Mapping[str, Optional[str]]]),
            ("repository_credentials", Optional[str]),
            ("server_health_check", Optional[Mapping[str, Any]]),
        ],
    )
):
    """Encapsulates configuration that can be applied to an ECS task running Dagster code."""

    def __new__(
        cls,
        secrets: Optional[Sequence[Any]] = None,
        secrets_tags: Optional[Sequence[str]] = None,
        env_vars: Optional[Sequence[str]] = None,
        task_definition_arn: Optional[str] = None,
        container_name: Optional[str] = None,
        server_resources: Optional[Mapping[str, str]] = None,
        run_resources: Optional[Mapping[str, str]] = None,
        task_role_arn: Optional[str] = None,
        execution_role_arn: Optional[str] = None,
        runtime_platform: Optional[Mapping[str, Any]] = None,
        mount_points: Optional[Sequence[Mapping[str, Any]]] = None,
        volumes: Optional[Sequence[Mapping[str, Any]]] = None,
        server_sidecar_containers: Optional[Sequence[Mapping[str, Any]]] = None,
        run_sidecar_containers: Optional[Sequence[Mapping[str, Any]]] = None,
        server_ecs_tags: Optional[Sequence[Mapping[str, Optional[str]]]] = None,
        run_ecs_tags: Optional[Sequence[Mapping[str, Optional[str]]]] = None,
        repository_credentials: Optional[str] = None,
        server_health_check: Optional[Mapping[str, Any]] = None,
    ):
        return super().__new__(
            cls,
            secrets=check.opt_sequence_param(secrets, "secrets"),
            secrets_tags=check.opt_sequence_param(secrets_tags, "secrets_tags"),
            env_vars=check.opt_sequence_param(env_vars, "env_vars"),
            task_definition_arn=check.opt_str_param(task_definition_arn, "task_definition_arn"),
            container_name=check.opt_str_param(container_name, "container_name"),
            server_resources=check.opt_mapping_param(server_resources, "server_resources"),
            run_resources=check.opt_mapping_param(run_resources, "run_resources"),
            task_role_arn=check.opt_str_param(task_role_arn, "task_role_arn"),
            execution_role_arn=check.opt_str_param(execution_role_arn, "execution_role_arn"),
            runtime_platform=check.opt_mapping_param(
                runtime_platform, "runtime_platform", key_type=str
            ),
            mount_points=check.opt_sequence_param(mount_points, "mount_points"),
            volumes=check.opt_sequence_param(volumes, "volumes"),
            server_sidecar_containers=check.opt_sequence_param(
                server_sidecar_containers, "server_sidecar_containers"
            ),
            run_sidecar_containers=check.opt_sequence_param(
                run_sidecar_containers, "run_sidecar_containers"
            ),
            server_ecs_tags=check.opt_sequence_param(server_ecs_tags, "server_ecs_tags"),
            run_ecs_tags=check.opt_sequence_param(run_ecs_tags, "run_tags"),
            repository_credentials=check.opt_str_param(
                repository_credentials, "repository_credentials"
            ),
            server_health_check=check.opt_mapping_param(server_health_check, "server_health_check"),
        )

    def merge(self, other: "EcsContainerContext") -> "EcsContainerContext":
        return EcsContainerContext(
            secrets=[*other.secrets, *self.secrets],
            secrets_tags=[*other.secrets_tags, *self.secrets_tags],
            env_vars=[*other.env_vars, *self.env_vars],
            task_definition_arn=other.task_definition_arn or self.task_definition_arn,
            container_name=other.container_name or self.container_name,
            server_resources={**self.server_resources, **other.server_resources},
            run_resources={**self.run_resources, **other.run_resources},
            task_role_arn=other.task_role_arn or self.task_role_arn,
            execution_role_arn=other.execution_role_arn or self.execution_role_arn,
            runtime_platform=other.runtime_platform or self.runtime_platform,
            mount_points=[*other.mount_points, *self.mount_points],
            volumes=[*other.volumes, *self.volumes],
            server_sidecar_containers=[
                *other.server_sidecar_containers,
                *self.server_sidecar_containers,
            ],
            run_sidecar_containers=[*other.run_sidecar_containers, *self.run_sidecar_containers],
            server_ecs_tags=[*other.server_ecs_tags, *self.server_ecs_tags],
            run_ecs_tags=[*other.run_ecs_tags, *self.run_ecs_tags],
            repository_credentials=other.repository_credentials or self.repository_credentials,
            server_health_check=other.server_health_check or self.server_health_check,
        )

    def get_secrets_dict(self, secrets_manager) -> Mapping[str, str]:
        return {
            **(get_tagged_secrets(secrets_manager, self.secrets_tags) if self.secrets_tags else {}),
            **{secret["name"]: secret["valueFrom"] for secret in self.secrets},
        }

    def get_environment_dict(self) -> Mapping[str, str]:
        parsed_env_var_tuples = [parse_env_var(env_var) for env_var in self.env_vars]
        return {env_var_tuple[0]: env_var_tuple[1] for env_var_tuple in parsed_env_var_tuples}

    @staticmethod
    def create_for_run(dagster_run: DagsterRun, run_launcher: Optional["EcsRunLauncher[Any]"]):
        context = EcsContainerContext()
        if run_launcher:
            context = context.merge(
                EcsContainerContext(
                    secrets=run_launcher.secrets,
                    secrets_tags=run_launcher.secrets_tags,
                    env_vars=run_launcher.env_vars,
                    task_definition_arn=run_launcher.task_definition,  # run launcher converts this from short name to ARN in constructor
                    run_resources=run_launcher.run_resources,
                    task_role_arn=run_launcher.task_role_arn,
                    execution_role_arn=run_launcher.execution_role_arn,
                    runtime_platform=run_launcher.runtime_platform,
                    mount_points=run_launcher.mount_points,
                    volumes=run_launcher.volumes,
                    run_sidecar_containers=run_launcher.run_sidecar_containers,
                    run_ecs_tags=run_launcher.run_ecs_tags,
                    repository_credentials=run_launcher.repository_credentials,
                )
            )

        run_container_context = (
            dagster_run.job_code_origin.repository_origin.container_context
            if dagster_run.job_code_origin
            else None
        )

        if not run_container_context:
            return context

        return context.merge(EcsContainerContext.create_from_config(run_container_context))

    @staticmethod
    def create_from_config(run_container_context) -> "EcsContainerContext":
        processed_shared_container_context = process_shared_container_context_config(
            run_container_context or {}
        )
        shared_container_context = EcsContainerContext(
            env_vars=processed_shared_container_context.get("env_vars", [])
        )

        run_ecs_container_context = (
            run_container_context.get("ecs", {}) if run_container_context else {}
        )

        if not run_ecs_container_context:
            return shared_container_context

        processed_container_context = process_config(
            ECS_CONTAINER_CONTEXT_SCHEMA, run_ecs_container_context
        )

        if not processed_container_context.success:
            raise DagsterInvalidConfigError(
                "Errors while parsing ECS container context",
                processed_container_context.errors,
                run_ecs_container_context,
            )

        processed_context_value = cast(Mapping[str, Any], processed_container_context.value)

        return shared_container_context.merge(
            EcsContainerContext(
                secrets=processed_context_value.get("secrets"),
                secrets_tags=processed_context_value.get("secrets_tags"),
                env_vars=processed_context_value.get("env_vars"),
                task_definition_arn=processed_context_value.get("task_definition_arn"),
                container_name=processed_context_value.get("container_name"),
                server_resources=processed_context_value.get("server_resources"),
                run_resources=processed_context_value.get("run_resources"),
                task_role_arn=processed_context_value.get("task_role_arn"),
                execution_role_arn=processed_context_value.get("execution_role_arn"),
                runtime_platform=processed_context_value.get("runtime_platform"),
                mount_points=processed_context_value.get("mount_points"),
                volumes=processed_context_value.get("volumes"),
                server_sidecar_containers=processed_context_value.get("server_sidecar_containers"),
                run_sidecar_containers=processed_context_value.get("run_sidecar_containers"),
                server_ecs_tags=processed_context_value.get("server_ecs_tags"),
                run_ecs_tags=processed_context_value.get("run_ecs_tags"),
                repository_credentials=processed_context_value.get("repository_credentials"),
                server_health_check=processed_context_value.get("server_health_check"),
            )
        )



import time
from typing import TYPE_CHECKING, Any, Literal, Optional, Union, cast

import boto3
import dagster._check as check
from dagster import PipesClient
from dagster._annotations import preview, public
from dagster._core.definitions.metadata import RawMetadataMapping
from dagster._core.definitions.resource_annotation import TreatAsResourceParam
from dagster._core.errors import DagsterExecutionInterruptedError
from dagster._core.execution.context.asset_execution_context import AssetExecutionContext
from dagster._core.execution.context.compute import OpExecutionContext
from dagster._core.pipes.client import (
    PipesClientCompletedInvocation,
    PipesContextInjector,
    PipesMessageReader,
)
from dagster._core.pipes.context import PipesSession
from dagster._core.pipes.utils import PipesEnvContextInjector, open_pipes_session

from dagster_aws.pipes.clients.utils import WaiterConfig, emr_inject_pipes_env_vars
from dagster_aws.pipes.message_readers import PipesCloudWatchMessageReader

if TYPE_CHECKING:
    from mypy_boto3_emr_containers.client import EMRContainersClient
    from mypy_boto3_emr_containers.type_defs import (
        DescribeJobRunResponseTypeDef,
        StartJobRunRequestRequestTypeDef,
        StartJobRunResponseTypeDef,
    )

AWS_SERVICE_NAME = "EMR Containers"


@public
@preview
class PipesEMRContainersClient(PipesClient, TreatAsResourceParam):
    """A pipes client for running workloads on AWS EMR Containers.

    Args:
        client (Optional[boto3.client]): The boto3 AWS EMR containers client used to interact with AWS EMR Containers.
        context_injector (Optional[PipesContextInjector]): A context injector to use to inject
            context into AWS EMR Containers workload. Defaults to :py:class:`PipesEnvContextInjector`.
        message_reader (Optional[PipesMessageReader]): A message reader to use to read messages
            from the AWS EMR Containers workload. It's recommended to use :py:class:`PipesS3MessageReader`.
        forward_termination (bool): Whether to cancel the AWS EMR Containers workload if the Dagster process receives a termination signal.
        pipes_params_bootstrap_method (Literal["args", "env"]): The method to use to inject parameters into the AWS EMR Containers workload. Defaults to "args".
        waiter_config (Optional[WaiterConfig]): Optional waiter configuration to use. Defaults to 70 days (Delay: 6, MaxAttempts: 1000000).
    """

    AWS_SERVICE_NAME = AWS_SERVICE_NAME

    def __init__(
        self,
        client: Optional["EMRContainersClient"] = None,
        context_injector: Optional[PipesContextInjector] = None,
        message_reader: Optional[PipesMessageReader] = None,
        forward_termination: bool = True,
        pipes_params_bootstrap_method: Literal["args", "env"] = "env",
        waiter_config: Optional[WaiterConfig] = None,
    ):
        self._client = client or boto3.client("emr-containers")
        self._context_injector = context_injector or PipesEnvContextInjector()
        self._message_reader = message_reader or PipesCloudWatchMessageReader()
        self.forward_termination = check.bool_param(forward_termination, "forward_termination")
        self.pipes_params_bootstrap_method = pipes_params_bootstrap_method
        self.waiter_config = waiter_config or WaiterConfig(Delay=6, MaxAttempts=1000000)

    @property
    def client(self) -> "EMRContainersClient":
        return self._client

    @property
    def context_injector(self) -> PipesContextInjector:
        return self._context_injector

    @property
    def message_reader(self) -> PipesMessageReader:
        return self._message_reader

    @classmethod
    def _is_dagster_maintained(cls) -> bool:
        return True

    @public
    def run(
        self,
        *,
        context: Union[OpExecutionContext, AssetExecutionContext],
        start_job_run_params: "StartJobRunRequestRequestTypeDef",
        extras: Optional[dict[str, Any]] = None,
    ) -> PipesClientCompletedInvocation:
        """Run a workload on AWS EMR Containers, enriched with the pipes protocol.

        Args:
            context (Union[OpExecutionContext, AssetExecutionContext]): The context of the currently executing Dagster op or asset.
            params (dict): Parameters for the ``start_job_run`` boto3 AWS EMR Containers client call.
                See `Boto3 EMR Containers API Documentation <https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/emr-containers/client/start_job_run.html>`_
            extras (Optional[Dict[str, Any]]): Additional information to pass to the Pipes session in the external process.

        Returns:
            PipesClientCompletedInvocation: Wrapper containing results reported by the external
            process.
        """
        with open_pipes_session(
            context=context,
            message_reader=self.message_reader,
            context_injector=self.context_injector,
            extras=extras,
        ) as session:
            start_job_run_params = self._enrich_start_params(context, session, start_job_run_params)
            start_response = self._start(context, start_job_run_params)
            try:
                completion_response = self._wait_for_completion(context, start_response)
                context.log.info(f"[pipes] {self.AWS_SERVICE_NAME} workload is complete!")
                return PipesClientCompletedInvocation(
                    session, metadata=self._extract_dagster_metadata(completion_response)
                )

            except DagsterExecutionInterruptedError:
                if self.forward_termination:
                    context.log.warning(
                        f"[pipes] Dagster process interrupted! Will terminate external {self.AWS_SERVICE_NAME} workload."
                    )
                    self._terminate(context, start_response)
                raise

    def _enrich_start_params(
        self,
        context: Union[OpExecutionContext, AssetExecutionContext],
        session: PipesSession,
        params: "StartJobRunRequestRequestTypeDef",
    ) -> "StartJobRunRequestRequestTypeDef":
        # inject Dagster tags
        tags = params.get("tags", {})
        params["tags"] = {**tags, **session.default_remote_invocation_info}

        params["jobDriver"] = params.get("jobDriver", {})

        if self.pipes_params_bootstrap_method == "env":
            params["configurationOverrides"] = params.get("configurationOverrides", {})
            params["configurationOverrides"]["applicationConfiguration"] = params[
                "configurationOverrides"
            ].get("applicationConfiguration", [])
            # we can reuse the same method as in standard EMR
            # since configurations format is the same
            params["configurationOverrides"]["applicationConfiguration"] = (
                emr_inject_pipes_env_vars(
                    session,
                    params["configurationOverrides"]["applicationConfiguration"],
                    emr_flavor="containers",
                )
            )

        # the other option is sparkSqlJobDriver - in this case there won't be a remote Pipes session
        # and no Pipes messages will arrive from the job
        # but we can still run it and get the logs
        if spark_submit_job_driver := params["jobDriver"].get("sparkSubmitJobDriver"):
            if self.pipes_params_bootstrap_method == "args":
                spark_submit_job_driver["sparkSubmitParameters"] = spark_submit_job_driver.get(
                    "sparkSubmitParameters", ""
                )

                for key, value in session.get_bootstrap_cli_arguments().items():
                    spark_submit_job_driver["sparkSubmitParameters"] += f" {key} {value}"

            params["jobDriver"]["sparkSubmitJobDriver"] = spark_submit_job_driver

        return cast("StartJobRunRequestRequestTypeDef", params)

    def _start(
        self,
        context: Union[OpExecutionContext, AssetExecutionContext],
        params: "StartJobRunRequestRequestTypeDef",
    ) -> "StartJobRunResponseTypeDef":
        response = self.client.start_job_run(**params)

        virtual_cluster_id = response["virtualClusterId"]
        job_run_id = response["id"]

        context.log.info(
            f"[pipes] {self.AWS_SERVICE_NAME} job started with job_run_id {job_run_id} on virtual cluster {virtual_cluster_id}."
        )

        return response

    def _wait_for_completion(
        self,
        context: Union[OpExecutionContext, AssetExecutionContext],
        start_response: "StartJobRunResponseTypeDef",
    ) -> "DescribeJobRunResponseTypeDef":
        job_run_id = start_response["id"]
        virtual_cluster_id = start_response["virtualClusterId"]

        # TODO: use a native boto3 waiter instead of a while loop
        # once it's available (it does not exist at the time of writing)
        attempts = 0
        while attempts < self.waiter_config.get("MaxAttempts", 1000000):
            response = self.client.describe_job_run(
                id=job_run_id, virtualClusterId=virtual_cluster_id
            )
            state = response["jobRun"].get("state")

            if state in ["COMPLETED", "FAILED", "CANCELLED"]:
                break

            time.sleep(self.waiter_config.get("Delay", 6))

        if state in ["FAILED", "CANCELLED"]:
            raise RuntimeError(
                f"EMR Containers job run {job_run_id} failed with state {state}. Reason: {response['jobRun'].get('failureReason')}, details: {response['jobRun'].get('stateDetails')}"
            )

        return self.client.describe_job_run(virtualClusterId=virtual_cluster_id, id=job_run_id)

    def _extract_dagster_metadata(
        self, response: "DescribeJobRunResponseTypeDef"
    ) -> RawMetadataMapping:
        metadata: RawMetadataMapping = {}

        metadata["AWS EMR Containers Virtual Cluster ID"] = response["jobRun"].get(
            "virtualClusterId"
        )
        metadata["AWS EMR Containers Job Run ID"] = response["jobRun"].get("id")

        # TODO: it would be great to add a url to EMR Studio page for this run
        # such urls look like: https://es-638xhdetxum2td9nc3a45evmn.emrstudio-prod.eu-north-1.amazonaws.com/#/containers-applications/00fm4oe0607u5a1d
        # but we need to get the Studio ID from the application_id
        # which is not possible with the current AWS API

        return metadata

    def _terminate(
        self,
        context: Union[OpExecutionContext, AssetExecutionContext],
        start_response: "StartJobRunResponseTypeDef",
    ):
        virtual_cluster_id = start_response["virtualClusterId"]
        job_run_id = start_response["id"]
        context.log.info(f"[pipes] Terminating {self.AWS_SERVICE_NAME} job run {job_run_id}")
        self.client.cancel_job_run(virtualClusterId=virtual_cluster_id, id=job_run_id)
        context.log.info(f"[pipes] {self.AWS_SERVICE_NAME} job run {job_run_id} terminated.")



from dagster_aws.ecs.executor import ecs_executor as ecs_executor
from dagster_aws.ecs.launcher import EcsRunLauncher as EcsRunLauncher
from dagster_aws.ecs.tasks import EcsEventualConsistencyTimeout as EcsEventualConsistencyTimeout
